/*!
 ************************************************************************
 * \file adaptive_quantization.c
 *
 * \brief
 *    adaptive quantization matrix selection routines
 *
 * \author
 *  Akiyuki Tanizawa       <akiyuki.tanizawa@toshiba.co.jp>           \n
 *
 * This software may be used for research purposes only.
 * TOSHIBA would also appreciate that any technical report, conference or
 * journal paper which uses the software includes one of the following references:
 *    - A. Tanizawa, T. Chujoh, "Adaptive Quantization Matrix Selection",
 *                              VCEG contribution D.266, Geneva, November 2006.
 *    - A. Tanizawa, T. Chujoh, "Simulation results of AQMS on KTA software",
 *                              VCEG contribution VCEG-AC07, Klagenfurt, July 2006.
 *    - A. Tanizawa, T. Chujoh, "Adaptive Quantization Matrix Selection on KTA software",
 *                              VCEG contribution VCEG-AD06, Hnagzhou, Oct. 2006.
 ************************************************************************
 */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <assert.h>
#include <limits.h>
#include <string.h>
#include "global.h"
#include "parset.h"
#include "adaptive_quantization.h"
#include "memalloc.h"
#include "malloc.h"
#include "defines.h"
#include "mbuffer.h"
#include "image.h"

#ifdef ADAPTIVE_QUANTIZATION
extern int quant_org[16];
extern int quant8_org[64];
extern const int dequant_coef[6][4][4];
static const int dequant_coef8[6][8][8] = 
{
  {
    {20,  19, 25, 19, 20, 19, 25, 19},
    {19,  18, 24, 18, 19, 18, 24, 18},
    {25,  24, 32, 24, 25, 24, 32, 24},
    {19,  18, 24, 18, 19, 18, 24, 18},
    {20,  19, 25, 19, 20, 19, 25, 19},
    {19,  18, 24, 18, 19, 18, 24, 18},
    {25,  24, 32, 24, 25, 24, 32, 24},
    {19,  18, 24, 18, 19, 18, 24, 18}
  },
  {
    {22,  21, 28, 21, 22, 21, 28, 21},
    {21,  19, 26, 19, 21, 19, 26, 19},
    {28,  26, 35, 26, 28, 26, 35, 26},
    {21,  19, 26, 19, 21, 19, 26, 19},
    {22,  21, 28, 21, 22, 21, 28, 21},
    {21,  19, 26, 19, 21, 19, 26, 19},
    {28,  26, 35, 26, 28, 26, 35, 26},
    {21,  19, 26, 19, 21, 19, 26, 19}
  },
  {
    {26,  24, 33, 24, 26, 24, 33, 24},
    {24,  23, 31, 23, 24, 23, 31, 23},
    {33,  31, 42, 31, 33, 31, 42, 31},
    {24,  23, 31, 23, 24, 23, 31, 23},
    {26,  24, 33, 24, 26, 24, 33, 24},
    {24,  23, 31, 23, 24, 23, 31, 23},
    {33,  31, 42, 31, 33, 31, 42, 31},
    {24,  23, 31, 23, 24, 23, 31, 23}
  },
  {
    {28,  26, 35, 26, 28, 26, 35, 26},
    {26,  25, 33, 25, 26, 25, 33, 25},
    {35,  33, 45, 33, 35, 33, 45, 33},
    {26,  25, 33, 25, 26, 25, 33, 25},
    {28,  26, 35, 26, 28, 26, 35, 26},
    {26,  25, 33, 25, 26, 25, 33, 25},
    {35,  33, 45, 33, 35, 33, 45, 33},
    {26,  25, 33, 25, 26, 25, 33, 25}
  },
  {
    {32,  30, 40, 30, 32, 30, 40, 30},
    {30,  28, 38, 28, 30, 28, 38, 28},
    {40,  38, 51, 38, 40, 38, 51, 38},
    {30,  28, 38, 28, 30, 28, 38, 28},
    {32,  30, 40, 30, 32, 30, 40, 30},
    {30,  28, 38, 28, 30, 28, 38, 28},
    {40,  38, 51, 38, 40, 38, 51, 38},
    {30,  28, 38, 28, 30, 28, 38, 28}
  },
  {
    {36,  34, 46, 34, 36, 34, 46, 34},
    {34,  32, 43, 32, 34, 32, 43, 32},
    {46,  43, 58, 43, 46, 43, 58, 43},
    {34,  32, 43, 32, 34, 32, 43, 32},
    {36,  34, 46, 34, 36, 34, 46, 34},
    {34,  32, 43, 32, 34, 32, 43, 32},
    {46,  43, 58, 43, 46, 43, 58, 43},
    {34,  32, 43, 32, 34, 32, 43, 32}
  }
  
};

//-------------for header.c
/*!
 ************************************************************************
 * \brief
 *    For generating the modeling quantization matrix
 *
 ************************************************************************
 */
void CreateModelingMatrix(int type)
{
  int i;

  gpModelingMatrixType->gdwMatrixGradient   = img->slice_scaling_model_param0[type];
  gpModelingMatrixType->gdwMatrixDCgain     = img->slice_scaling_model_param1[type];

  if(type)
  {
    for(i=0;i<6;i++)
      ScalingList4x4AQMS[i] = active_pps->ScalingList4x4AQMS[i];
    for(i=0;i<2;i++)
      ScalingList8x8AQMS[i] = active_pps->ScalingList8x8AQMS[i];
  }
  else
  {
    for(i=0;i<6;i++)
      ScalingList4x4[i] = active_pps->ScalingList4x4[i];
    for(i=0;i<2;i++)
      ScalingList8x8[i] = active_pps->ScalingList8x8[i];
  }

  SetAutoScalingListData(type);
}

//-------------for block.c
/*!
 ************************************************************************
 * \brief
 *    For assigning the modeling quantization matrix
 *
 ************************************************************************
 */
void AssignQuantParamForAQMS(pic_parameter_set_rbsp_t* pps)
{
  int i;

  if( img->slice_adaptive_quant_matrix_flag )
  {
    for(i=0; i<8; i++)
    {
      if(i<6)
        qmatrix[i] = pps->ScalingList4x4[i];
      else
        qmatrix[i] = pps->ScalingList8x8[i-6];
    }

    for(i=0; i<8; i++)
    {
      qmatrixAQMS[i] = (i<6) ? pps->ScalingList4x4AQMS[i] : pps->ScalingList8x8AQMS[i-6];
    }
  }
}
/*!
 ************************************************************************
 * \brief
 *    For calculating inverse-level scale for AQMS
 *
 ************************************************************************
 */
void CalculateQuantParamForAQMS()
{
  int i,j,k,temp;
  if(img->slice_adaptive_quant_matrix_flag)
  {
    for(k=0; k<6; k++)
      for(j=0; j<4; j++)
        for(i=0; i<4; i++)
        {
          temp = (i<<2)+j;
          InvLevelScale4x4Luma_Intra_AQMS[k][j][i]      = dequant_coef[k][j][i]*qmatrixAQMS[0][temp];
          InvLevelScale4x4Chroma_Intra_AQMS[0][k][j][i] = dequant_coef[k][j][i]*qmatrixAQMS[1][temp];
          InvLevelScale4x4Chroma_Intra_AQMS[1][k][j][i] = dequant_coef[k][j][i]*qmatrixAQMS[2][temp];

          InvLevelScale4x4Luma_Inter_AQMS[k][j][i]      = dequant_coef[k][j][i]*qmatrixAQMS[3][temp];
          InvLevelScale4x4Chroma_Inter_AQMS[0][k][j][i] = dequant_coef[k][j][i]*qmatrixAQMS[4][temp];
          InvLevelScale4x4Chroma_Inter_AQMS[1][k][j][i] = dequant_coef[k][j][i]*qmatrixAQMS[5][temp];
        }
  }
}

//-------------for parset.c
/*!
 ************************************************************************
 * \brief
 *    For generating the modeling quantization matrix accroding to InternalFlag
 *
 ************************************************************************
 */
void SetAutoScalingListData(int InternalFlag)
{
  int pA = 0, pB = 0,  pT = 0;

  pA = ((gpModelingMatrixType->gdwMatrixGradient)-32);
  pB = (gpModelingMatrixType->gdwMatrixDCgain);

  // 4x4 block
  CreateScalingList4x4(pT, pA, pB, InternalFlag);
  // 8x8 block
  CreateScalingList8x8(pT, pA, pB, InternalFlag);
}
/*!
 ************************************************************************
 * \brief
 *    For generating the modeling quantization matrix at 4x4 block
 *
 ************************************************************************
 */
void CreateScalingList4x4(int pT, int pA, int pB, int InternalFlag)
{
  int i,j,r=0;
  unsigned int QM4[16];

  // 4x4 Matrix
  switch ( pT )
  {
    case 0:
      {
        for(j=0;j<4;j++){
          for(i=0;i<4;i++){
            r=i+j;
            QM4[i+(j<<2)]=Clip3(1,255,(((pA*r)+8)>>4) + pB);
          }
        }
      }
      break;
    default:
      {
        printf("ERROR\n");
        break;
      }
  }

  if(InternalFlag)
  {
    for(i=0;i<16;i++)
      for(j=0;j<6;j++)
        ScalingList4x4AQMS[j][i] = Clip3(1,255,QM4[i]);
  }
  else
  {
    for(i=0;i<16;i++)
      for(j=0;j<6;j++)
        ScalingList4x4[j][i] = Clip3(1,255,QM4[i]);
  }
}
/*!
 ************************************************************************
 * \brief
 *    For generating the modeling quantization matrix at 8x8 block
 *
 ************************************************************************
 */
void CreateScalingList8x8(int pT, int pA, int pB, int InternalFlag)
{
  int          i,j,r=0;
  unsigned int QM[64];

  // 8x8 Matrix
  switch ( pT )
  {
    case 0:
    {
      for(j=0;j<8;j++){
        for(i=0;i<8;i++){
          r=i+j;
          QM[i+j*8]=Clip3(1,255,( ((pA*r+16)>>5 ) + pB) );
        }
      }
      break;
    }
    default:
    {
      printf("ERROR\n");
      break;
    }
  }

  if(InternalFlag)
  {
    for(i=0;i<64;i++)
      for(j=0;j<2;j++)
      ScalingList8x8AQMS[j][i] = Clip3(1,255,QM[i]);
  }
  else
  {
    for(i=0;i<64;i++)
      for(j=0;j<2;j++)
        ScalingList8x8[j][i] = Clip3(1,255,QM[i]);
  }
}

/*!
 ************************************************************************
 * \brief
 *    For calculating the inverse level scale
 *
 ************************************************************************
 */
void CalculateQuant8ParamForAQMS()
{
  int i,j,k,temp;
  if(img->slice_adaptive_quant_matrix_flag)
  {
    for(k=0; k<6; k++)
      for(j=0; j<8; j++)
        for(i=0; i<8; i++)
        {
          temp = (i<<3)+j;
          InvLevelScale8x8Luma_Intra_AQMS[k][j][i] = dequant_coef8[k][j][i]*qmatrixAQMS[6][temp];
          InvLevelScale8x8Luma_Inter_AQMS[k][j][i] = dequant_coef8[k][j][i]*qmatrixAQMS[7][temp];
        }
  }
}

#endif
